# -*- python -*-
# $RCSfile: skeletongroupmenu.py,v $
# $Revision: 1.30 $
# $Author: redwing $
# $Date: 2005/06/06 13:14:59 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# This file builds the menu commands that operate the methods
# of the various selectable-groups, corresponding to methods
# of GenericGroupSet class in skeletongroups.py

from oof2.SWIG.common import ooferror
from oof2.common import debug
from oof2.common import utils
from oof2.common.IO import mainmenu
from oof2.common.IO import oofmenu
from oof2.common.IO import parameter
from oof2.common.IO import whoville
from oof2.common.IO import automatic
from oof2.common.IO import reporter
from oof2.common.IO import xmlmenudump
from oof2.engine import skeletongroups
from oof2.engine.IO import skeletongroupparams
import types

AutomaticNameParameter = parameter.AutomaticNameParameter
NodeGroupParameter = skeletongroupparams.NodeGroupParameter
SegmentGroupParameter = skeletongroupparams.SegmentGroupParameter
ElementGroupParameter = skeletongroupparams.ElementGroupParameter

nodegroupmenu = oofmenu.OOFMenuItem(
    "NodeGroup",
    cli_only=1,
    help='Create and manage NodeGroups.',
    discussion="""<para>
    &nodegroups; provide a convenient way of operating on many &skel;
    &nodes; simultaneously.
    </para>""")

segmentgroupmenu = oofmenu.OOFMenuItem(
    "SegmentGroup",
    cli_only=1,
    help='Create and manage SegmentGroups.',
    discussion="""<para>
    &sgmtgroups; provide a convenient way of operating on many &skel;
    &sgmts; simultaneously.
    </para>""")

elementgroupmenu = oofmenu.OOFMenuItem(
    "ElementGroup",
    cli_only=1,
    help='Create and manage ElementGroups.',
    discussion="""<para>
    &elemgroups; provide a convenient way of operating on many &skel;
    &elems; simultaneously.
    </para>""")


mainmenu.OOF.addItem(nodegroupmenu)
mainmenu.OOF.addItem(segmentgroupmenu)
mainmenu.OOF.addItem(elementgroupmenu)


########################################################################
# Generic callbacks usable by all classes of selectables.  Menu items
# must have a data item which is a string corresponding to the name
# of the selection set in the skeletoncontext, e.g. "nodegroups",
# so that the menuitem callback can retrieve the right groupset.

# Make a new group name known.

def _new_group(menuitem, skeleton, name):
    skelc = whoville.getClass('Skeleton')[skeleton]
    groupset = getattr(skelc, menuitem.data)
    if groupset.isGroup(name):
        raise ooferror.ErrUserError("Group %s already exists." % group)
    else:
        groupset.addGroup(name)


# Remove a group.

def _remove_group(menuitem, skeleton, group):
    skelc = whoville.getClass('Skeleton')[skeleton]
    groupset = getattr(skelc, menuitem.data)
    if not groupset.isGroup(group):
        raise ooferror.ErrUserError("Group %s does not exist." % group)
    else:
        groupset.removeGroup(group)


# Copy an existing group.
def _copy_group(menuitem, skeleton, group, new_name):
    skelc = whoville.getClass('Skeleton')[skeleton]
    groupset = getattr(skelc, menuitem.data)
    if groupset.isGroup(new_name):
        raise ooferror.ErrUserError("Group %s already exists." % new_name)
    if groupset.isGroup(group):
        groupset.copyGroup(group, new_name)


# Add the current selection to the named group, creating
# the group if required.

def _add_selection_to_group(menuitem, skeleton, group):
    skelc = whoville.getClass('Skeleton')[skeleton]
    groupset = getattr(skelc, menuitem.data)
    if not groupset.isGroup(group):
        groupset.addGroup(group)
    groupset.addSelectionToGroup(group)

# Remove the currently-selected objects from the named group.
def _remove_selection_from_group(menuitem, skeleton, group):
    skelc = whoville.getClass('Skeleton')[skeleton]
    groupset = getattr(skelc, menuitem.data)
    # if not groupset.isGroup(group):
    #     groupset.removeGroup(group)
    groupset.removeSelectionFromGroup(group)

    
# Rename an existing group.
def _rename_group(menuitem, skeleton, group, new_name):
    skelc = whoville.getClass('Skeleton')[skeleton]
    groupset = getattr(skelc, menuitem.data)
    groupset.renameGroup(group, new_name)
##     if groupset.isGroup(new_name):
##         raise ooferror.ErrUserError("Group %s already exists." % new_name)
##     if groupset.isGroup(group):
##         groupset.renameGroup(group, new_name)
        

# Clear a group.

def _clear_group(menuitem, skeleton, group):
    skelc = whoville.getClass('Skeleton')[skeleton]
    groupset = getattr(skelc, menuitem.data)
    if not groupset.isGroup(group):
        raise ooferror.ErrUserError("Group %s does not exist." % group)
    groupset.clearGroup(group)


# Query a node group.
def _query_node_group(menuitem, skeleton, group):
    skelc = whoville.getClass('Skeleton')[skeleton]
    members = skelc.nodegroups.get_group(group)
    reporter.report(">>> ", len(members), " nodes")

def _query_sgmt_group(menuitem, skeleton, group):
    skelc = whoville.getClass('Skeleton')[skeleton]
    members = skelc.segmentgroups.get_group(group)
    reporter.report(">>> ", len(members), " segments")

def _query_elem_group(menuitem, skeleton, group):
    skelc = whoville.getClass('Skeleton')[skeleton]
    members = skelc.elementgroups.get_group(group)
    area = 0.0
    homog = 0.0
    for element in members:
        area += element.area()
        homog += element.homogeneity(skelc.getObject())
    if len(members):
        homog /= len(members)
    reporter.report(">>> ", len(members), " elements,", " area: ", area,
                    ", average homogeneity: ", homog)

#######################################

class GroupNameResolver:
    # Callable object to be used as the resolver for the
    # AutomaticNameParameter for group names.  Requires derived
    # classes that define getBasename().
    def __init__(self, attrname):
        # attrname is the name of the SkeletonContext attribute that
        # holds the groupset, eg, "nodegroups".
        self.attrname = attrname
    def __call__(self, param, startname):
        skelname = param.group['skeleton'].value
        skelcontext = whoville.getClass('Skeleton')[skelname]
        groupset = getattr(skelcontext, self.attrname)
        return utils.uniqueName(self.getBasename(param, startname),
                                groupset.allGroups())

class NewGroupNameResolver(GroupNameResolver):
    # Gets the default name from the constructor arguments.
    def __init__(self, basename, attrname):
        GroupNameResolver.__init__(self, attrname)
        # basename is the defaultname to be used for the groups.
        self.basename = basename
    def getBasename(self, param, startname):
        if param.automatic():
            return self.basename
        return startname

class CopyGroupNameResolver(GroupNameResolver):
    # Gets the default name from the object being copied (or renamed).
    def getBasename(self, param, startname):
        if param.automatic():
            return param.group['group'].value
        return startname
    
#######################################    
# Node-specific menu items:

new = oofmenu.OOFMenuItem(
    "New_Group",
    cli_only=1,
    callback=_new_group,
    params=parameter.ParameterGroup(
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    AutomaticNameParameter("name", value=automatic.automatic,
                           resolver=NewGroupNameResolver("nodegroup",
                                                         "nodegroups"),
                           tip="Name of the group to be created.")),
    help="Create a new node group.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/newnodegroup.xml')
    )

new.data = "nodegroups"

nodegroupmenu.addItem(new)

rename = oofmenu.OOFMenuItem(
    "Rename_Group", cli_only=1,
    callback=_rename_group,
    params=parameter.ParameterGroup(
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    NodeGroupParameter("group", tip="Name of the group to be renamed."),
    parameter.StringParameter("new_name", "",
                              tip="New name for the node group")),
    help="Rename a node group.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/renamenodegroup.xml')
    )
                             

rename.data = "nodegroups"

nodegroupmenu.addItem(rename)


copy = oofmenu.OOFMenuItem(
    "Copy_Group",
    cli_only=1,
    callback=_copy_group,
    params=parameter.ParameterGroup(
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    NodeGroupParameter("group", tip="Name of the group to be copied."),
    AutomaticNameParameter("new_name",
                           resolver=CopyGroupNameResolver("nodegroups"),
                           value=automatic.automatic,
                           tip="Name for the copy.")),
    help="Make a copy of a node group.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/copynodegroup.xml')
    )

copy.data = "nodegroups"

nodegroupmenu.addItem(copy)


remove = oofmenu.OOFMenuItem(
    "Delete_Group",
    cli_only=1,
    callback=_remove_group,
    params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    NodeGroupParameter("group", tip="Name of the group to be deleted.")],
    help="Remove a node group.",
    discussion="""<para>
    Delete the given &nodegroup; from the &skel;. The &nodes;
    themselves are not affected.
    </para>""")

remove.data = "nodegroups"

nodegroupmenu.addItem(remove)



addselect = oofmenu.OOFMenuItem(
    "Add_to_Group",
    cli_only=1,
    callback=_add_selection_to_group,
    params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    NodeGroupParameter("group", tip="Name of the group.")],
    help="Add selected nodes to the group.",
    discussion="""<para>
    Add the currently selected &skel; &nodes; to the given &nodegroup;.
    </para>""") 

addselect.data = "nodegroups"

nodegroupmenu.addItem(addselect)


removeselect = oofmenu.OOFMenuItem(
    "Remove_from_Group",
    cli_only=1,
    callback=_remove_selection_from_group,
    params=[whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                                  tip=parameter.emptyTipString),
            NodeGroupParameter("group", tip="Name of the group")],
    help="Remove selected nodes from the group.",
    discussion="""<para>
    Remove the currently selected &skel; &nodes; from the given &nodegroup;.
    </para>""") 

removeselect.data = "nodegroups"

nodegroupmenu.addItem(removeselect)


clear = oofmenu.OOFMenuItem(
    "Clear_Group", cli_only=1,
    callback=_clear_group,
    params=[whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                                  tip=parameter.emptyTipString),
            NodeGroupParameter("group", tip="Name of the group.")],
    help="Clear a node group.",
    discussion="""<para>
    Remove all &nodes; from the given &nodegroup;.  The group itself
    will <emphasis>not</emphasis> be deleted, but will be emptied.
    </para>""")

clear.data = "nodegroups"

nodegroupmenu.addItem(clear)

query = oofmenu.OOFMenuItem(
    "Query_Group",
    cli_only=1,
    callback=_query_node_group,
    params=[whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                                  tip=parameter.emptyTipString),
            NodeGroupParameter("group", tip="Name of the group.")],
    help="Query a node group.",
    discussion="<para>Print information about the given &nodegroup;.</para>") 

query.data = "nodegroups"

nodegroupmenu.addItem(query)

#########################################################
# Segments.

new = oofmenu.OOFMenuItem(
    "New_Group",
    cli_only=1,
    callback=_new_group,
    params=parameter.ParameterGroup(
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    AutomaticNameParameter("name",
                           value=automatic.automatic,
                           resolver=NewGroupNameResolver("segmentgroup",
                                                         "segmentgroups"),
                           tip="Name of the group.")),
    help="Create a new segment group.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/newsegmentgroup.xml')
    )

new.data = "segmentgroups"

segmentgroupmenu.addItem(new)


rename = oofmenu.OOFMenuItem(
    "Rename_Group",
    cli_only=1,
    callback=_rename_group,
    params=parameter.ParameterGroup(
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    SegmentGroupParameter("group", tip="Name of the group to be renamed."),
    parameter.StringParameter("new_name", "",
                              tip="New name for the segment group")),
    help="Rename a segment group.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/renamesegmentgroup.xml')
    )

rename.data = "segmentgroups"

segmentgroupmenu.addItem(rename)


copy = oofmenu.OOFMenuItem(
    "Copy_Group",
    cli_only=1,
    callback=_copy_group,
    params=parameter.ParameterGroup(
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    SegmentGroupParameter("group", tip="Name of the group to be copied."),
    AutomaticNameParameter("new_name",
                           resolver=CopyGroupNameResolver("segmentgroups"),
                           value=automatic.automatic,
                           tip="New name for the copy.")),
    help="Copy a segment group.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/copysegmentgroup.xml')
    )

copy.data = "segmentgroups"

segmentgroupmenu.addItem(copy)


remove = oofmenu.OOFMenuItem(
    "Delete_Group",
    cli_only=1,
    callback=_remove_group,
    params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    SegmentGroupParameter("group", tip="Name of the group.")],
    help="Remove a segment group.",
    discussion="""<para>
    Delete the given &sgmtgroup; from the &skel;.  The &sgmts;
    themselves are not affected.
    </para>""")

remove.data = "segmentgroups"

segmentgroupmenu.addItem(remove)



addselect = oofmenu.OOFMenuItem(
    "Add_to_Group",
    cli_only=1,
    callback=_add_selection_to_group,
    params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    SegmentGroupParameter("group", tip="Name of the group.")],
    help="Add selected segments to the group.",
    discussion="""<para>
    Add the currently selected &skel; &sgmts; to the given &sgmtgroup;.
    </para>""") 

addselect.data = "segmentgroups"

segmentgroupmenu.addItem(addselect)


removeselect = oofmenu.OOFMenuItem(
    "Remove_from_Group",
    cli_only=1,
    callback=_remove_selection_from_group,
    params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    SegmentGroupParameter("group", tip="Name of the group.")],
    help="Remove selected segments from the group.",
    discussion="""<para>
    Remove the currently selected &skel; &sgmts; from the given &sgmtgroup;.
    </para>""")

removeselect.data = "segmentgroups"

segmentgroupmenu.addItem(removeselect)


clear = oofmenu.OOFMenuItem(
    "Clear_Group",
    cli_only=1,
    callback=_clear_group,
    params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    SegmentGroupParameter("group", tip="Name of the group.")],
    help="Clear a segment group.",
    discussion="""<para>
    Remove all &sgmts; from the given &sgmtgroup;.  The group itself
    will <emphasis>not</emphasis> be deleted, but will be emptied.
    </para>""")

clear.data = "segmentgroups"

segmentgroupmenu.addItem(clear)


query = oofmenu.OOFMenuItem(
    "Query_Group",
    cli_only=1,
    callback=_query_sgmt_group,
    params=[whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                                  tip=parameter.emptyTipString),
            SegmentGroupParameter("group", tip="Name of the group.")],
    help="Query a segment group.",
    discussion="<para>Print information about the given &sgmtgroup;.</para>") 

query.data = "segmentgroups"

segmentgroupmenu.addItem(query)

########################################################
# Elements

new = oofmenu.OOFMenuItem(
    "New_Group",
    cli_only=1,
    callback=_new_group,
    params=parameter.ParameterGroup(
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    AutomaticNameParameter("name",
                           value=automatic.automatic,
                           resolver=NewGroupNameResolver("elementgroup",
                                                         "elementgroups"),
                           tip="Name for the element group to be created.")),
    help="Create a new element group.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/newelementgroup.xml')
    )

new.data = "elementgroups"


elementgroupmenu.addItem(new)


rename = oofmenu.OOFMenuItem(
    "Rename_Group",
    cli_only=1,
    callback=_rename_group,
    params=parameter.ParameterGroup(
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    ElementGroupParameter("group", tip="Name of the group to be renamed."),
    parameter.StringParameter("new_name", "",
                              tip="New name for the element group")),
    help="Rename an element group.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/renameelementgroup.xml')
    )

rename.data = "elementgroups"

elementgroupmenu.addItem(rename)


copy = oofmenu.OOFMenuItem(
    "Copy_Group",
    cli_only=1,
    callback=_copy_group,
    params=parameter.ParameterGroup(
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    ElementGroupParameter("group", tip="Name of the group to be copied."),
    AutomaticNameParameter("new_name",
                           resolver=CopyGroupNameResolver("elementgroups"),
                           value=automatic.automatic,
                           tip="Name for the copy.")),
    help="Make a copy of an element group.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/copyelementgroup.xml')
    )

copy.data = "elementgroups"

elementgroupmenu.addItem(copy)


remove = oofmenu.OOFMenuItem(
    "Delete_Group",
    cli_only=1,
    callback=_remove_group,
    params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    ElementGroupParameter("group", tip="Name of the group.")],
    help="Remove an element group.",
    discussion="""<para>
    Delete the given &elemgroup; from the &skel;. The &elems;
    themselves are not affected.
    </para>""") 

remove.data = "elementgroups"

elementgroupmenu.addItem(remove)



addselect = oofmenu.OOFMenuItem(
    "Add_to_Group",
    cli_only=1,
    callback=_add_selection_to_group,
    params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    ElementGroupParameter("group", tip="Name of the group.")],
    help="Add selected elements to the group.",
    discussion="""<para>
    Add the currently selected &skel; &elems; to the given &elemgroup;.
    </para>""")

addselect.data = "elementgroups"

elementgroupmenu.addItem(addselect)


removeselect = oofmenu.OOFMenuItem(
    "Remove_from_Group",
    cli_only=1,
    callback=_remove_selection_from_group,
    params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    ElementGroupParameter("group", tip="Name of the group.")],
    help="Remove selected elements from the group.",
    discussion="""<para>
    Remove the currently selected &skel; &elems; from the given &elemgroup;.
    </para>""")

removeselect.data = "elementgroups"

elementgroupmenu.addItem(removeselect)


clear = oofmenu.OOFMenuItem(
    "Clear_Group",
    cli_only=1,
    callback=_clear_group,
    params=[whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                                  tip=parameter.emptyTipString),
            ElementGroupParameter("group", tip="Name of the group.")],
    help="Clear an element group.",
    discussion="""<para>
    Remove all &elems; from the given &elemgroup;.  The group itself
    will <emphasis>not</emphasis> be deleted, but will be emptied.
    </para>""")

clear.data = "elementgroups"

elementgroupmenu.addItem(clear)

query = oofmenu.OOFMenuItem(
    "Query_Group",
    cli_only=1,
    callback=_query_elem_group,
    params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    ElementGroupParameter("group", tip="Name of the group.")],
    help="Query an element group.",
    discussion="<para>Print information about the given &elemgroup;. </para>")

query.data = "elementgroups"

elementgroupmenu.addItem(query)
