# -*- python -*-
# $RCSfile: mesh.py,v $
# $Revision: 1.106 $
# $Author: langer $
# $Date: 2005/06/20 15:52:46 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# The Mesh class wraps the C++ FEMesh class, associating a Skeleton
# and miscellaneous bookkeeping information with it.  Mesh itself is a
# Who class, which means it can participate in the layer-display
# process.

from oof2.SWIG.common import devel
from oof2.SWIG.common import lock
from oof2.SWIG.common import ooferror
from oof2.SWIG.common import switchboard
from oof2.SWIG.common import timestamp
from oof2.SWIG.engine import equation
from oof2.SWIG.engine import field
from oof2.SWIG.engine import flux
from oof2.SWIG.engine import masterelement
from oof2.common import debug
from oof2.common import enum
from oof2.common import ringbuffer
from oof2.common import utils
from oof2.common.IO import parameter
from oof2.common.IO import whoville
import copy
import string


# Helper class for keeping track of the cross-sections -- remembers
# what the most-recently-added one was by remembering the order.
class CrossSectionSet:
    def __init__(self):
        self.data = {}
        self.namelist = []
        self._selection = None
        self.timestamp = timestamp.TimeStamp()
    def new(self, name, cs):
        self.data[name]=cs
        self.namelist.append(name)
        self.timestamp.increment()
        self._selection = name           # always select new cross section
        switchboard.notify("cross sections changed")
    def remove(self, name):
        del self.data[name]
        self.namelist.remove(name)
        self.timestamp.increment()
        if self._selection == name:
            if self.namelist:
                self._selection = self.namelist[-1]
            else:
                self._selection = None
        switchboard.notify("cross sections changed")
    def all_names(self):
        return self.namelist
    def __getitem__(self, key):
        return self.data[key]
    def selected(self):
        try:
            return self.data[self._selection]
        except KeyError:
            return None
    def selectedName(self):
        return self._selection
    def select(self, name):
        self._selection = name
        self.timestamp.increment()
        switchboard.notify("cross sections changed")
    def deselect(self):
        self._selection = None
        self.timestamp.increment()
        switchboard.notify("cross sections changed")
    def rename(self, oldname, newname):
        cs = self.data[oldname]
        newname = utils.uniqueName(newname, self.namelist, exclude=oldname)
        self.namelist[self.namelist.index(oldname)] = newname
        del self.data[oldname]
        self.data[newname] = cs
        if oldname == self._selection:
            self._selection = newname
        switchboard.notify("cross sections changed")
        return newname
    def replace(self, name, cross_section):
        self.data[name] = cross_section
        self.timestamp.increment()
        switchboard.notify("cross sections changed")
        
class StatusItem:
    def __init__(self, state=0, message=""):
        self.state = state
        self.msg = message
    def __eq__(self, val):
        return self.state == val
    def __repr__(self):
        return self.msg

######################

class FEMeshMapOrder(enum.EnumClass(
    ('1', 'linear map functions'),
    ('2', 'quadratic map functions'))):
    tip = "Order of mapping functions."
    discussion = """
    <para>Order of mapping functions.
    </para>"""                    

class FEMeshShapeOrder(enum.EnumClass(
    ('1', 'linear shape functions'),
    ('2', 'quadratic shape functions'))):
    tip = "Order of shape functions."
    discussion = """
    <para>Order of shape functions.
    </para>"""                    

######################

skel_stack_size = 50
# TODO: RingBuffer is not a terribly good choice for this.
class SkeletonBuffer(ringbuffer.RingBuffer):
    def makeCurrent(self, index):  # make the given index the current position.
        self.currentpos = index
        return self.current()

    def isFirstInLine(self, skel):
        return self.data.index(skel)==0
    
class Mesh(whoville.Who):
    def __init__(self, name, classname, femesh,
                 skeleton=None, elementdict=None,
                 materialfactory=None):
        whoville.Who.__init__(self, name, classname, femesh)
        # Passed-in femesh is returned by self.getObject()
        self.initializers = {}
        # have to straighten is_current flags.  Timestamps
        # handled in parent class.
        self.mesh_is_current = 1
        self.solution_is_current = 0
        # bdy bdy.
        self.bdyconditions = {}
        # Skeletons will be stored in the ring buffer
        self.skeleton_buffer = SkeletonBuffer(skel_stack_size)
        self.pushSkeleton(skeleton)
        # This will add skeleton to the buffer and so some more
        
        self.elementdict = elementdict
        self.materialfactory = materialfactory
        femesh.set_parent_mesh(self)
        # Cross-section stuff
        self.cross_sections = CrossSectionSet()

        # Timestamps.
        # MeshChanged is incremented when a field or equation is added
        # or removed.  Boundarieschanged is incremented whenever a
        # boundary or boundary condition is added or removed.
        # Materialschanged is incremented for material operations, and
        # solutiontimestamp is incremented at solution-time.
        self.meshchanged = timestamp.TimeStamp() # last mesh modification
        self.boundarieschanged = timestamp.TimeStamp() # last bdy change
        self.materialschanged = timestamp.TimeStamp()
        self.solutiontimestamp = timestamp.TimeStamp() # last "solve" command
        self.solutiontimestamp.backdate()

        # Status
        self.solverStatus = None
        self.solverMessage = ""
        self.solver = None
        self.meshSymmetry = "Symmetry Unknown"
        # Materials inconsistent -- recompute each time a relevant
        # material changes or a material is added or deleted.

        # Materials defined -- recompute each time a material is added
        # or deleted.

        # Mesh Solution -- self.solutiontimestamp > self.meshchanged and
        # self.solutiontimestamp > self.materialschanged

    def pushSkeleton(self, skeleton):
        self.skeleton_buffer.push(skeleton)
        self.setSkeleton(skeleton)  # set the current skeleton
        skeleton.add_mesh(self)

    def getTimeStamp(self, gfxwindow):
        return max(self.meshchanged, self.materialschanged,
                   self.solutiontimestamp, self.boundarieschanged)

    def femesh(self):
        return self.getObject()

    def size(self):
        return self.skeleton.size()

    # utility function from "femesh"
    def nelements(self):
        return self.getObject().nelements()

    def nnodes(self):
        return self.getObject().nnodes()
    
    # Required for compatibility with Display objects which work for
    # both Mesh and Skeleton who classes.  
    def mesh(self):
        return self.getObject()

    def getSkeleton(self):  # returns the current skeleton.
        return self.skeleton

    def setSkeleton(self, skeleton):
        self.skeleton = skeleton

    def setFemesh(self, femesh):
        self._obj = femesh  # replace the old one with the new one
        self._obj.set_parent_mesh(self)

    def getMicrostructure(self):
        return self.skeleton.MS

    def getMasterElementType(self, nsides):
        try:
            return self.elementdict[nsides].name()
        except KeyError:
            return '---'

    def destroy(self):
        global meshes
        self.skeleton.remove_mesh(self)
        del self.skeleton               # may not be necessary but doesn't hurt
        # Remove internal circular references in the FEMesh.  These
        # attributes are added in femesh_.py.
        self.getObject().destroy()  # includes te next four commented lines.
##        self.getObject().clear_parent_mesh()
##        del self.getObject().boundaries
##        del self.getObject().edgesets
##        del self.getObject().nodesets
        # Remove from the enclosing WhoClass.  This nulls out parent's "._obj".
        meshes.remove(self.path())

    def initialize_field(self, field, initializer):
        self.initializers[field] = initializer
        initializer.apply(self.getObject(), field)
        self.solution_is_current = 0
        switchboard.notify("redraw")

    def all_fields(self):
        # Because Fields aren't PythonExportable objects,
        # FEMesh.all_fields() returns a list of CompoundField objects
        # without the correct subclass associations.  However, field
        # names must be unique, so we can get the correct field with
        # the full class hierarchy by looking it up in the list
        # maintained by the field module.
        fields = self.getObject().all_fields()
        return [field.allFields[f.name()] for f in fields]

    def all_equations(self):
        return self.getObject().all_equations()

    def all_fluxes(self):
        fluxes = self.getObject().all_fluxes()
        return [flux.allFluxes[f.name()] for f in fluxes]

    def allCrossSectionNames(self):
        return self.cross_sections.all_names()

    def addCrossSection(self, name, cs):
        self.cross_sections.new(name, cs)

    def removeCrossSection(self, name):
        self.cross_sections.remove(name)

    def uniqueCSName(self, name):
        return utils.uniqueName(name, self.cross_sections.all_names())

##    def lastCS(self):
##        return self.cross_sections.last()

##    def lastCSName(self):
##        return self.cross_sections.last_name()

    def selectedCS(self):
        return self.cross_sections.selected()

    def selectedCSName(self):
        return self.cross_sections.selectedName()

    def getCrossSection(self, name):
        return self.cross_sections[name]

    def selectCrossSection(self, name):
        self.cross_sections.select(name)

    def deselectCrossSection(self):
        self.cross_sections.deselect()

    def renameCrossSection(self, oldname, newname):
        actualnewname = self.cross_sections.rename(oldname, newname)
        ## This switchboard call is here, rather than in
        ## CrossSectionSet.rename, because it needs to send the Mesh
        ## as an argument.
        switchboard.notify(('cross section renamed', self),
                           oldname, actualnewname)

    def replaceCrossSection(self, name, cross_section):
        self.cross_sections.replace(name, cross_section)
        
    # Boundary retrieval functions -- the interface needs to know
    # the names.

    def getBoundary(self, name):
        return self.getObject().boundaries[name]

    # Names of nontrivial boundaries -- does not return a complete
    # list, only returns the names of those objects with size > 0.
    # This is the meaning of the "Finite" modifier.
    def boundaryNames(self):
        names= self.getObject().getFiniteBoundaryNames()
        return names

    def pointBoundaryNames(self):
        return self.getObject().getFinitePointBdyNames()

    def edgeBoundaryNames(self):
        return self.getObject().getFiniteEdgeBdyNames()
    
    # Post-construction-time changes to the boundaries can also happen.
    
    def newEdgeBoundary(self, name, skeleton_bdy):
        realmesh = self.getObject()
        realbdy = realmesh.newEdgeBoundary(name)
        for e in skeleton_bdy.edges:
            skelel = e.segment.getElements()[0]
            realel = realmesh.getElement(skelel.meshindex)

            edge_nodes = e.get_nodes()
            realn0 = realmesh.getNode( edge_nodes[0].meshindex )
            realn1 = realmesh.getNode( edge_nodes[1].meshindex )
            realbdy.addEdge(realel.getBndyEdge(realn0,realn1))
        switchboard.notify("mesh boundaries changed", self)



    def newPointBoundary(self, name, skeleton_bdy):
        realmesh = self.getObject()
        realbndy = realmesh.newPointBoundary(name)
        for node in skeleton_bdy.nodes:
            realbndy.addNode(realmesh.getNode(node.meshindex))
        switchboard.notify("mesh boundaries changed", self)
            
            
    def removeBoundary(self, name):
        for (bcname, bc) in self.allBoundaryConds():
            if bc.boundary==name:
                bc.remove_from_mesh()
                self.bdys_changed()
        self.getObject().removeBoundary(name)
        switchboard.notify("mesh boundaries changed", self)
        

    def renameBoundary(self, oldname, newname):
        for (bcname, bc) in self.allBoundaryConds():
            if bc.boundary==oldname:
                bc.boundary=newname
        self.getObject().renameBoundary(oldname, newname)
        switchboard.notify("mesh boundaries changed", self)

    def replacePointBoundary(self, name, new_skel_boundary):
        del self.getObject().boundaries[name]
        self.newPointBoundary(name, new_skel_boundary)
        for (bcname, bc) in self.allBoundaryConds():
            if bc.boundary==name:
                bc.boundary_obj = self.getObject().boundaries[name]
                bc.boundary_obj.addCondition(bc)
                self.bdys_changed()
        switchboard.notify("mesh boundaries changed", self)
        

    def replaceEdgeBoundary(self, name, new_skel_boundary):
        del self.getObject().boundaries[name]
        self.newEdgeBoundary(name, new_skel_boundary)
        for (bcname, bc) in self.allBoundaryConds():
            if bc.boundary==name:
                bc.boundary_obj = self.getObject().boundaries[name]
                bc.boundary_obj.addCondition(bc)
                self.bdys_changed()
        self.boundarieschanged.increment()
        switchboard.notify("mesh boundaries changed", self)

    
    # Boundary condition manipulation functions.  The "promotion"
    # argument indicates whether or not this modification involves the
    # promotion of a Float BC to a Dirichlet BC.  In this case, no
    # real change is occurring in the mesh, and in particular, you
    # don't want to draw it any differently, so "self.changed" is not
    # called.  Recipients of the "boundary conditions changed" signal
    # also get the "promotion" argument, and should behave
    # appropriately.
    def addBdyCondition(self, name, bc, promotion=0):
        try:
            self.bdyconditions[name]
        except KeyError:
            self.bdyconditions[name]=bc
            switchboard.notify("boundary conditions changed",
                               self, name, promotion)
        else:
            raise ooferror.ErrSetupError(
                "Duplicate boundary condition name %s." % name )
        if not promotion:
            self.bdys_changed()
        
    def uniqueBCName(self, name):
        return utils.uniqueName(name, self.bdyconditions.keys())

    def renameBdyCondition(self, oldname, newname):
        try:
            bc = self.bdyconditions[oldname]
        except KeyError:
            raise ooferror.ErrSetupError("Can't find boundary condition %s!"
                                         % oldname)
        newname = utils.uniqueName(newname, self.allBndyCondNames(),
                                   exclude = oldname)
        if newname != oldname:
            del self.bdyconditions[oldname]
            self.bdyconditions[newname] = bc
        # Zero at the end is because this is not a promotion.
        self.boundarieschanged.increment()
        switchboard.notify("boundary conditions changed",
                           self, newname, 0)

    def rmBdyCondition(self, name, promotion=0):
        self.bdyconditions[name].disconnect()
        del self.bdyconditions[name]
        switchboard.notify("boundary conditions changed",
                           self, name, promotion)
        if not promotion:
            self.bdys_changed()
        
    def allBoundaryConds(self):
        return self.bdyconditions.items()

    def allBndyCondNames(self):
        return self.bdyconditions.keys()

    def getBdyCondition(self, name):
        return self.bdyconditions[name]
    
    # Make a copy of the boundary condition, but on the indicated
    # boundary.  Caller is responsible for adding it to the mesh,
    # because that operation should remain at the menu level.
    def copyBdyCondition(self, bc, boundary):
        return self.bdyconditions[bc].copy(boundary)

    # This function must be called after fields or equations are added
    # to or removed from this femesh -- BCs which have become
    # incomputable will shut themselves off, and those which have
    # become computable will switch themselves on.
    def autoenableBCs(self):
        for (name, bc) in self.allBoundaryConds():
            bc.auto_enable()
        # Mildly hacky -- third argument is BC to select, fourth is
        # promotion status -- in this case, we don't want to change
        # the selection, and we aren't promoting.  This function could
        # have it's own signal.
        switchboard.notify("boundary conditions changed", self, None, 0)

    def refreshMaterials(self):
        self.reserve()
        self.begin_writing()
        try:
            self.getObject().refreshMaterials(self.skeleton)
            self.materialschanged.increment()
        finally:
            self.end_writing()
            self.cancel_reservation()
        switchboard.notify("mesh changed", self)

    def materialsConsistent(self):
        # check_materials returns a list of inconsistent materials
        return (not self.getObject().check_materials()) and \
               (not self.getObject().check_equations())

    def materialsNonlinear(self):
        return self.getObject().nonlinear_materials()

    # For field/equation/materials changes, which will require
    # recomputation of the stiffness matrix.
    def changed(self):
        self.meshchanged.increment()    # timestamp
        self.getObject().requirePrecompute()
        switchboard.notify("mesh changed", self)

    # For boundary/boundary-condition changes, which require
    # the solver to be re-run, but which do not require the stiffness
    # matrix to be rebuilt.
    def bdys_changed(self):
        self.boundarieschanged.increment()
        switchboard.notify("mesh changed", self)

    def get_initializer(self, field):
        try:
            return self.initializers[field]
        except KeyError:
            return None

    def enclosingElement(self, point):
        skel = self.getSkeleton()
        skelel = skel.enclosingElement(point)
        meshel = self.getObject().getElement(skelel.meshindex)
        return meshel

    def setSolverStatus(self, success, solver, message):
        self.solverStatus = success or 'failed'
        self.solver = solver
        self.solverMessage = message or ''
    def make_linear_system(self, force=0):
        r=self.materialsNonlinear()
        if (self.meshchanged > self.solutiontimestamp) or \
               (self.materialschanged > self.solutiontimestamp) or \
               self.materialsNonlinear() or force:
            debug.fmsg("Rebuilding stiffness matrix.")
            self.getObject().make_linear_system()
            debug.fmsg("after Rebuilding stiffness matrix.")
    def get_symmetry_state(self):
        return self.meshSymmetry
    def set_symmetry_state(self, state):
        if state>0:
            ## print "setting symmetric state"
            self.meshSymmetry = "Symmetric"
        else:
            self.meshSymmetry = "Unsymmetric"
            ## print "setting UNsymmetric state"

    # Function to assess the well-posedness of the current state.
    # Returns None if the problem is well-posed, and returns a message
    # string indicating the problem if it's not well-posed.
    def check_materials(self):
        status = None
        badmatls = self.getObject().check_materials()
        badeqns = self.getObject().check_equations()
        if badmatls or badeqns:
            status= "Ill-posed problem!\n"
            if badmatls:
                status += "Badly defined material%s!\n" % \
                          ("s"*(len(badmatls)>1))
                for matl in badmatls:
                    status += string.join(
                        matl.consistency_messages(), '\n')
            if badeqns:
                status += "Equation%s with no flux contributions!\n" % \
                          ("s"*(len(badeqns)>1))
                for eqn in badeqns:
                    status += eqn.name() + "\n"
        return status
            
    def status(self):
        if self.solutiontimestamp > \
               max(self.materialschanged,
                   self.meshchanged,
                   self.boundarieschanged):
            debug.fmsg('meshSymmetry=', self.meshSymmetry)
            debug.fmsg('solverStatus=', self.solverStatus)
            debug.fmsg('solverMessage=', self.solverMessage)
            status = "Stiffness Matrix " + self.meshSymmetry + "\n" + \
                     "Solution " + self.solverStatus + "\n" + \
                     "Solver: " + `self.solver` + '\n' + \
                     "Message: " + self.solverMessage + '\n'
        else:
            material_error = self.check_materials()
            if material_error:
                status = material_error + "\n" # Extra new-line for clarity.
            else:
                status = 'Unsolved' + "\n" + \
                         "Stiffness Matrix " + self.meshSymmetry + "\n"
        if devel.devel():
            status += "time index = " + str(self.getObject().current_time()) + "\n"
        return status

    def solved(self):
        self.solution_is_current = 1
        switchboard.notify("mesh changed", self)

    def has_solution(self):
        return self.solution_is_current
            
        
from oof2.engine.skeletoncontext import skeletonContexts

meshes = whoville.WhoClass(
    'Mesh',
    ordering=300,
    parentClass=skeletonContexts,
    instanceClass=Mesh,
    proxyClasses=['<topmost>','<contourable>'])

utils.OOFdefine('meshes', meshes)
