# -*- python -*-
# $RCSfile: relaxation.py,v $
# $Revision: 1.28 $
# $Author: langer $
# $Date: 2005/06/10 19:57:13 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from oof2.SWIG.common import devel
from oof2.SWIG.common import switchboard
from oof2.SWIG.common.IO import progressbar
from oof2.SWIG.engine import equation
from oof2.SWIG.engine import field
from oof2.SWIG.engine import masterelement
from oof2.SWIG.engine import material
from oof2.SWIG.engine import preconditioner
from oof2.SWIG.engine import solver
from oof2.SWIG.engine import solverdriver
from oof2.SWIG.engine import steppingscheme
from oof2.SWIG.engine.property.elasticity.iso import iso
from oof2.SWIG.engine.property.heatconductivity import heatconductivity
from oof2.SWIG.engine.property.orientation import orientation
from oof2.SWIG.engine.property.skeletonrelaxationrate import skeletonrelaxationrate
from oof2.common import debug
from oof2.common import primitives
from oof2.common import registeredclass
from oof2.common.IO import parameter
from oof2.common.IO import reporter
from oof2.common.IO import xmlmenudump
from oof2.engine import bdycondition
from oof2.engine import fieldinit
from oof2.engine import profile
from oof2.engine import skeletonmodifier
from oof2.engine.IO import meshmenu
import oof2.engine.mesh
import random


# Make an instance of the elasticity property at start up time with
# default parameter values.  It can't be done when Relax is called
# because the parameters may have changed by then.
stiffness = iso.IsotropicElasticity()

class Relax(skeletonmodifier.SkeletonModifier):
    intro = "Preparing to relax skeleton. Please be cool...  "
    header = "relaxing skeleton: "
    exit_message = "relaxation done "
    def __init__(self, alpha, gamma, iterations):
        self.alpha = alpha
        self.gamma = gamma
        self.iterations = iterations
        self.count = 0
        self.solver_converged = 1
        self.meshname = None

        ## create material
        self.skeletonMaterial = material.Material('__bone_marrow__')
        ## SkeletonRelaxationRateTensor is the PropertyRegistration for the
        ## SkeletonRelaxationRate property.
        beta_parameter = skeletonrelaxationrate.SkeletonRelaxationRateTensor.getParameter('beta')
        beta_parameter.value = self.alpha
        alpha_parameter = skeletonrelaxationrate.SkeletonRelaxationRateTensor.getParameter('alpha')
        alpha_parameter.value = self.gamma
        skelRelRate = skeletonrelaxationrate.SkeletonRelaxationRateTensor()
        ## assign properties for material
        self.skeletonMaterial.addProperty(stiffness)
        self.skeletonMaterial.addProperty(skelRelRate)
        
    def goodToGo(self):
        ## self.iter_number +=1
        if self.count >= self.iterations or self.solver_converged != 1:
            return 0
        return 1
    def updateIteration(self):
        self.count +=1
    def set_fakematerial(self, skelelement, skeleton):
        return self.skeletonMaterial
    def create_mesh(self, context):
        ## context is a skeleton context
        ## setup element types
        edict = {}
        # get linear isoparametric master elements
        edict[3] = masterelement.getMasterElementDict()['T3_3']
        edict[4] = masterelement.getMasterElementDict()['Q4_4']
        skel = context.getObject()
        ## returns a Mesh (Who) object
        self.meshname = context.path() + ":__internal_mesh__"
        femesh = skel.femesh(1, edict, self.set_fakematerial)
        meshcontext = oof2.engine.mesh.meshes.add(self.meshname, femesh,
                                             skeleton=skel, elementdict=edict)
        femesh.addExplicitMaterial(self.skeletonMaterial)
            
        femesh.set_microstructure(skel.MS)
        return meshcontext
    def delete_mesh(self, mesh):
        mesh.reserve()
        mesh.begin_writing()
        mesh.destroy()
        
    def define_fields(self, femesh): ## femesh is an femesh
        femesh.define_field(field.getCompoundField('Displacement'))

        ## also activate the fields
        femesh.activate_field(field.getCompoundField('Displacement'))
        field.getCompoundField('Displacement').set_in_plane(femesh, 1)
        
        
    def activate_equations(self, femesh): ## femesh is an femesh
        femesh.activate_equation(equation.getEquation('Force_Balance'))
    def set_boundary_conditions(self, mesh):
        ## here, mesh is a Mesh (Who) object
        
        ## left boundary
        self.leftBoundaryCondition = \
                                   bdycondition.DirichletBC(field.getCompoundField('Displacement'),
                                                            'x',
                                                            equation.getEquation('Force_Balance'),
                                                            'x',
                                                            profile.ConstantProfile(0),
                                                            'left'
                                                            )
        self.leftBoundaryCondition.add_to_mesh('left', mesh.path())
        
        ## right boundary
        self.rightBoundaryCondition = \
                                   bdycondition.DirichletBC(field.getCompoundField('Displacement'),
                                                            'x',
                                                            equation.getEquation('Force_Balance'),
                                                            'x',
                                                            profile.ConstantProfile(0),
                                                            'right'
                                                            )
        self.rightBoundaryCondition.add_to_mesh('right', mesh.path())
        
        ## top boundary
        self.topBoundaryCondition = \
                                   bdycondition.DirichletBC(field.getCompoundField('Displacement'),
                                                            'y',
                                                            equation.getEquation('Force_Balance'),
                                                            'y',
                                                            profile.ConstantProfile(0),
                                                            'top'
                                                            )
        self.topBoundaryCondition.add_to_mesh('top', mesh.path())
        
        ## bottom boundary
        self.bottomBoundaryCondition = \
                                   bdycondition.DirichletBC(field.getCompoundField('Displacement'),
                                                            'y',
                                                            equation.getEquation('Force_Balance'),
                                                            'y',
                                                            profile.ConstantProfile(0),
                                                            'bottom'
                                                            )
        self.bottomBoundaryCondition.add_to_mesh('bottom', mesh.path())
    def update_node_positions(self, skeleton, mesh):
        ## mesh is a Mesh Who object
        femesh = mesh.getObject()
        dispField = field.getCompoundField('Displacement')
        nodes = skeleton.nodes
        for node in nodes:
            realnode = femesh.getNode(node.meshindex)
            dx = dispField.value(femesh, realnode, 0)
            dy = dispField.value(femesh, realnode, 1)
            skeleton.moveNodeBy(node, primitives.Point(dx, dy))

    def apply(self, oldskeleton, context):
        pBar = progressbar.getProgress()
        pBar.set_message(self.intro)
        return oldskeleton.deputyCopy()
    def initialize_fields(self, mesh):
        initializer = fieldinit.ConstTwoVectorFieldInit(cx=0.0,cy=0.0)
        meshmenu.initField(self, self.meshname, field.getCompoundField('Displacement'), initializer)
    def postProcess(self, context):
        ## This function first creates a mesh with custom-made properties,
        ## then assigns "temporary" properties to pixels
        ## and specifies BCs and equations.
        ## Next, iterates for the solution using the specified solver,
        ## accepts all node moves and redraws skeleton.
        ## It repeats these steps until iteration criteria has been met.
        ## Finally, the (temporary) mesh and the rest  of the temporary
        ## objects are cleaned up.

        ## create progress bar
        pBar = progressbar.getProgress()
        pBar.set_message(self.header)
        
        ## get skeleton and calculate energy
        skeleton = context.getObject()
        before = skeleton.energyTotal(self.alpha)
        self.count = 0

        while self.goodToGo():
        ## create temporary mesh (Who) object
            mesh = self.create_mesh(context) ## femesh is created and properties are assigned
            
        ## define active fields
            self.define_fields(mesh.femesh()) ## defines displacement and temperature fields
            mesh.changed()
        ## activate relevant equations
            self.activate_equations(mesh.femesh()) ## activates the mechanical balance equation
            mesh.changed()
            self.set_boundary_conditions(mesh) ## constrains the nodes on the boundaries to 
        ## only slide along the edge
        
            ## self.initialize_fields(mesh)
            ## context thread-safe area starts here
            context.begin_writing()
            try:
                self.coreProcess(mesh)
                ## coreProcess solves a linear system.
                if pBar.query_stop():
                    pBar.set_failure()
                    pBar.set_message("Failed")
                    break
                self.update_node_positions(skeleton, mesh)
                skeleton.timestamp.increment()
                switchboard.notify("skeleton nodes moved", context)
            finally:
                context.end_writing()
                switchboard.notify("redraw")
                self.delete_mesh(mesh)
            ## context thread-safe area ends here
        

            self.updateIteration() ## update iteration manager machinery
            ## update progress bar machinery
            if pBar.query_stop():
                pBar.set_failure()
                pBar.set_message("Failed")
                break
            else:
                pBar.set_progress(1.0*self.count/self.iterations)
                pBar.set_message("%s%d/%d"
                                 % (self.header, self.count, self.iterations))
        ## update progress bar machinery
        if pBar.query_stop(): 
            pBar.set_failure()
            pBar.set_message("Failed")
            return
        
        ## calculate total energy improvement, if any.
        after = skeleton.energyTotal(self.alpha)
        if before:
            rate = 100.0*(before-after)/before
        else:
            rate = 0.0
        diffE = after - before
        reporter.report("%s deltaE = %10.4e (%6.3f%%)"
                        % (self.exit_message, diffE, rate))
        
        
        
    def coreProcess(self, mesh, activenodes = None):
        precond = preconditioner.blockPreconditioner(preconditioner.iluPreconditioner())
        krylov_method = solver.gmresSolver(1000, 200, 1e-5, precond)
        ## actual_solver = solverdriver.NewtonDriver(krylov_method, 1e-5, 1)
        actual_solver = solverdriver.LinearDriver(krylov_method)
        actual_solver.apply(steppingscheme.NullScheme(), mesh, verbosity = 0, report_success = 0)
        self.solver_converged = krylov_method.status() == 0 ## and actual_solver.status

        
       
            

#################################################

registeredclass.Registration(
    'Relax',
    skeletonmodifier.SkeletonModifier,
    Relax,
    ordering=0.5,
    params=[
    skeletonmodifier.alphaParameter,
    parameter.FloatParameter(name = 'gamma', value = 0.5, tip='Node mobility'),
    parameter.IntParameter('iterations', value = 1, tip='number of steps')
    ],
    tip='Improve a skeleton by solving a finite element system where the properties are functions of the underlying homogeneity and shape',
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/relax.xml')
    )


# Old tip for gamma:
#    tip='Coefficient of mesh expansion or rate of expansion/contraction per inhomogeneity or per deviation from ideal shape. A small positive value is recommended for meshes where element edges want to be made coincide with pixel group edges.'
